#!/usr/bin/env python

import sys, getopt, os, errno, json, subprocess, time, codecs

fullpath = '/Library/Application Support/Microsoft/Defender/wdavstate'

print_id = False
print_diag = True
populate_release_ring = False

def get_ostype():
    cmd = 'uname -s'
    try:
        return subprocess.check_output(cmd, shell = True).strip()
    except:
        print("OS not detected")
        sys.exit(2)

def get_uuid(ostype):
    if ostype == 'Darwin':
        uuid = subprocess.check_output('ioreg -ad2 -c IOPlatformExpertDevice | xmllint --xpath \'//key[.="IOPlatformUUID"]/following-sibling::*[1]/text()\' -', shell = True)
    else:
        uuid = subprocess.check_output('dmidecode -s system-uuid', shell = True)
        try:
            uuid = uuid.decode('utf-8').strip()
        except (UnicodeDecodeError, AttributeError):
            if os.path.exists('/sys/class/dmi/id/product_uuid'):
                f = open('/sys/class/dmi/id/product_uuid','r')
                uuid = f.readline().strip()
            else:
                pass
    return uuid

def usage():
    print("""Usage: %s [--printid] [--orgid='microsoft.com'] [--state=/path/to/state.json] [--releasering=ring] [--help]
    When no options are specified, prints out effective uuid and orgid
    --printid (or -t)     : Print just uuid (telemetry id)
    --state=path (or -s path): Use different state.json)
    --releasering=ring (or -r ring): Add specified ring to state
    --help (or -h): Print out this help page and exit
""" % sys.argv[0])

def get_orgid_from_plist(path):
    cmd = 'plutil -convert xml1 -o - "%s" | xmllint --xpath \'//key[.="OnboardingInfo"]/following-sibling::*[1]/text()\' - 2>/dev/null' % path

    try:
        json_payload = subprocess.check_output(cmd, shell = True)
        payload = json.loads(json_payload)
        json_body = payload['body']
        body = json.loads(json_body)
        orgid = body['orgId']

    except:
        orgid = ''
        pass

    return orgid

try:
    if os.getuid() != 0:
        print("Must be run as root.")
        sys.exit(1)

    opts, args = getopt.getopt(sys.argv[1:], 'hts:', ['help', 'printid', 'releasering=', 'state='])

    for k, v in opts:
        if k == '-t' or k == '--printid':
            print_id = True
            print_diag = False

        if k == '-s' or k == '--state':
            fullpath = v

        if k == '-r' or k == '--releasering':
            release_ring = v
            populate_release_ring = True
            print_diag = False

        if k == '-h' or k == '--help':
            usage()
            sys.exit(0)

except getopt.GetoptError as e:
    print(e)
    usage()
    sys.exit(2)

application_state = {}

try:
    fullpath = os.path.abspath(fullpath)
    dir = os.path.dirname(fullpath)
    os.makedirs(dir)
except OSError as e:
    if e.errno != errno.EEXIST:
        raise

if os.path.exists(fullpath):
    with open(fullpath, 'r') as fr:
        application_state = json.load(fr)

dirty = False

machine_guid = None

if application_state and 'machineGuid' in application_state: 
    machine_guid = application_state['machineGuid']

if not machine_guid:
    if print_diag:
        print('Generating telemetry id')

    ostype = get_ostype()
    uuid = get_uuid(ostype)

    if print_diag:
        print('Generated %s' % uuid)

    machine_guid = uuid
    dirty = True

if populate_release_ring:
    application_state['releaseRing'] = release_ring
    dirty = True

if print_id:
    # Must use sys.stdout.write to print w/o a newline for python2 and python3 compatibility
    sys.stdout.write(machine_guid)
    sys.stdout.flush()

if print_diag:
    print('uuid            : %s' % machine_guid)

    orgid_managed = get_orgid_from_plist('/Library/Managed Preferences/com.microsoft.wdav.atp.plist')
    orgid_manual = get_orgid_from_plist('/Library/Application Support/Microsoft/Defender/com.microsoft.wdav.atp.plist')

    orgid = ''
    if orgid_managed != '':
        orgid = orgid_managed
    elif orgid_manual != '':
        orgid = orgid_manual

    print('orgid           : %s' % orgid)
    
    if orgid_managed !='':
        print('orgid managed   : %s' % orgid_managed)
    
    if orgid_manual != '':
        print('orgid manual    : %s' % orgid_manual)

    try:
        with open(fullpath) as f:
            json_payload = json.load(f)
            effective_orgId = json_payload['enterpriseState']['orgId']
    except:
        effective_orgId = ''

    print('orgid effective : %s' % effective_orgId)

if dirty:
    if print_diag:
        print("Saving %s" % fullpath)

    application_state['machineGuid'] = machine_guid

    with codecs.open(fullpath, 'w', encoding="utf-8") as fw:
        json.dump(application_state, fw, sort_keys = True, indent = 4, ensure_ascii = False)
    os.chmod(fullpath, 0o600)

